/// эта библиотека была создана HarryTheCat
/// лицензия - CC BY SA

//
// train wheel update script
// upload wheel angle into state.wheelAngle
function updateWheelAngle(train, state) { 
	const angleLoop = 10000;

	// if (!train.shouldRenderDetail()) return 0;
	let speedMps = 20 * train.speed();
	let delta = Timing.delta();
	if (train.isReversed()) {
		state.wheelAngle -= speedMps * delta; 
	} else {
		state.wheelAngle += speedMps * delta; 
	}
	
	if (state.wheelAngle > angleLoop) state["wheelAngle"] = 0;
	if (state.wheelAngle < 0) state["wheelAngle"] = angleLoop;
}

function initWheelAngle(state) { 
	state.wheelAngle = 0;
}
// breaking script

function updateSpeedStates(train, state) { 
	state.isBreaking = train.speed() < state.oldSpeed;
	state.isAccelerating = train.speed() > state.oldSpeed;
	state.oldSpeed = train.speed();
}

function initSpeedStates(state) { 
	state.isBreaking = false;
	state.oldSpeed = 0.0;
	state.isAccelerating = false;
}

// train turning script
function updateTurnState(train, state) { 
	let absDelta =  Timing.delta() == 0 ? 0 : Math.abs(state.oldFirstCarTurn - train.lastCarRotation[0].y()) / (Timing.delta() * 50);
	
	if (train.isReversed()) {
		state.isTurningLeft = state.oldFirstCarTurn > train.lastCarRotation[0].y() && absDelta > 0.0025;
		state.isTurningRight = state.oldFirstCarTurn < train.lastCarRotation[0].y() && absDelta > 0.0025;
	} else {
		state.isTurningLeft = state.oldFirstCarTurn < train.lastCarRotation[0].y() && absDelta > 0.0025;
		state.isTurningRight = state.oldFirstCarTurn > train.lastCarRotation[0].y() && absDelta > 0.0025;
	}

	state.oldFirstCarTurn = train.lastCarRotation[0].y();
}
function initTurnState(train, state) {
	state.isTurningLeft = false;
	state.isTurningRight = false;
	state.oldFirstCarTurn = train.lastCarRotation[0].y();
}

// model to parts script
function uploadPartedModels(rawModels, removeRenderType, invertUV) { 
  result = {};
  for (it = rawModels.entrySet().iterator(); it.hasNext(); ) {
    entry = it.next();

    let jsStringKey = "" + entry.getKey(); 
    if (jsStringKey.includes("__int__")) {
    	entry.getValue().setAllRenderType("interior"); //включение интерьера
    	if (removeRenderType) jsStringKey = jsStringKey.replace("__int__", '');

    } else if (jsStringKey.includes("__light__")) {
    	entry.getValue().setAllRenderType("light"); //включение света
    	if (removeRenderType) jsStringKey = jsStringKey.replace("__light__", '');

    } else if (jsStringKey.includes("__window__")) {
    	entry.getValue().setAllRenderType("exteriortranslucent"); //
    	if (removeRenderType) jsStringKey = jsStringKey.replace("__window__", '');

    } else if (jsStringKey.includes("__windowint__")) {
    	entry.getValue().setAllRenderType("interiortranslucent"); //
    	if (removeRenderType) jsStringKey = jsStringKey.replace("__windowint__", '');

    } 

    print("registering " + jsStringKey + "...");

    entry.getValue().applyUVMirror(false, invertUV);
    result[jsStringKey] = ModelManager.uploadVertArrays(entry.getValue());
  }
  return result;
}

// roll script

function updateRoll(train, state, coef) {
	const mass = 0.17;

	if (Timing.delta() == 0) return;

	for (let i = 0; i < train.trainCars(); i++) {
		state.rolls[i] = coef * mass * (getCorrectDelta(train.lastCarRotation[i].y(), 
			state.oldYRotations[i])) * train.speed() / Timing.delta();

		if (train.isReversed()) state.rolls[i] *= -1;
		state.oldYRotations[i] = train.lastCarRotation[i].y();
	}
}

function initRoll(train, state){
	// let array = new Array(n).fill(0.0); // test
	state.rolls = [];
	state.oldYRotations = [];
	for (let i = 0; i < train.trainCars(); i++) {
		state.rolls[i] = 0.0;
		state.oldYRotations[i] = 0.0;
	}
}

///radius scripts
function initRadius(state){
	state.trainRads = [];
	state.radiusOldYRotations = [];
	for (let i = 0; i < train.trainCars(); i++) {
		state.trainRads[i] = 0.0;
		state.radiusOldYRotations[i] = 0.0;
	}
}

function updateRadius(train, state) {

	if (Timing.delta() == 0) return;

	for (let i = 0; i < train.trainCars(); i++) {
		let alpha = getCorrectDelta(train.lastCarRotation[i].y(), 
			state.radiusOldYRotations[i])
		let speed = train.speed() * 20;

		if (alpha == 0.0) {
			state.trainRads[i] = 0.0
		} else {
			state.trainRads[i] = (Timing.delta() * speed) / alpha
		}

		if (train.isReversed()) state.trainRads[i] *= -1;
		state.radiusOldYRotations[i] = train.lastCarRotation[i].y();
	}
}


///wiper scripts

function initWipers(state){
	state.wipeCoef = 0.0;
	state.wipeReverseMark = 0;
}

function updateWipers(train, state) {
	let pos = train.lastCarPosition[0];
	let isRaining = MinecraftClient.worldIsRainingAt(pos);
	if (isRaining || state.wipeCoef > 0) {
		state.wipeCoef += Timing.delta() / 1.5;
		if (state.wipeCoef > 1.0) {
			state.wipeCoef = 0.0;
			state.wipeReverseMark = train.isReversed();
		}
	}
}

/// car wheel scripts



function updateSteeringAngle(train, state) {
	if (Timing.delta() == 0.0) return;

	let delta = getCorrectDelta(train.lastCarRotation[0].y(), state.oldFirstCarTurnForSteering)
	if (delta == 0.0) {
		state.steeringAngle = 0.0;
		state.oldFirstCarTurnForSteering = train.lastCarRotation[0].y();
		return;
	}
	const coef = 0.0625;
	const l = 3.32;

	let radius = train.speed() * Timing.delta() / delta;
	let sin = clamp(coef * l / radius, -1.0, 1.0);

	if (train.isReversed()) sin *= -1;

	state.steeringAngle = Math.asin(sin);
	state.oldFirstCarTurnForSteering = train.lastCarRotation[0].y();
}

function updateSteeringAngleL(train, state, l) {
	if (Timing.delta() == 0.0) return;

	let delta = getCorrectDelta(train.lastCarRotation[0].y(), state.oldFirstCarTurnForSteering)
	if (delta == 0.0) {
		state.steeringAngle = 0.0;
		state.oldFirstCarTurnForSteering = train.lastCarRotation[0].y();
		return;
	}
	const coef = 0.0625;

	let radius = train.speed() * Timing.delta() / delta;
	let sin = clamp(coef * l / radius, -1.0, 1.0);

	if (train.isReversed()) sin *= -1;

	state.steeringAngle = Math.asin(sin);
	state.oldFirstCarTurnForSteering = train.lastCarRotation[0].y();
}


function initSteeringAngle(train, state) {
	state.steeringAngle = 0.0;
	state.oldFirstCarTurnForSteering = train.lastCarRotation[0].y();
	// тут oldFirstCarTurnForSteering и oldFirstCarTurn хранят одни и те же значения, 
	// но являются разными полями state. Думаю, не нужно просчитывать логику для использования одной переменной, т.к. это нагрузит код 
}
/// rocking script

function initRocking(train, state) {
	state.curvyRailCoef = 0.0;
	state.rockingZCoef = 0.0;
	state.rockingXCoef = 0.0;
}

function updateRocking(train, state, xCoef, zCoef, curvyXCoef, curvyZCoef) {
	const x = 0.05;
	const z = 0.05;
	const curvyspeed = 4.0;

	let railIndex = train.getRailIndex(train.railProgress(), true);
	let name = train.path()[railIndex].rail == null ? "" : train.path()[railIndex].rail.getModelKey();
	let isOnCurvyRail = name.contains("_curvy_");


	if (isOnCurvyRail) {
		state.curvyRailCoef += Timing.delta() / curvyspeed;
		if (state.curvyRailCoef > 1.0) state.curvyRailCoef = 1.0;
	} else {
		state.curvyRailCoef -= Timing.delta() / curvyspeed;
		if (state.curvyRailCoef < 0.0) state.curvyRailCoef = 0.0;
	}

	//лерпит значения
	let totalXcoef = xCoef + state.curvyRailCoef * (curvyXCoef - xCoef);
	let totalZcoef = zCoef + state.curvyRailCoef * (curvyZCoef - zCoef);

	let addX = Timing.delta() * train.speed() / (x * totalXcoef);
	let addZ = Timing.delta() * train.speed() / (z * totalZcoef);



	state.rockingXCoef += addX;
	state.rockingZCoef += addZ;

	if (state.rockingXCoef > 1.0) state.rockingXCoef = 0.0;
	if (state.rockingZCoef > 1.0) state.rockingZCoef = 0.0;
}

/// utils

function clamp(num, min, max) {
	if (num > max) return max;
	if (num < min) return min;
	return num;
}

function getCorrectDelta(angle1, angle2) {
  let delta = angle1 - angle2;
  if (Math.abs(delta) > Math.PI) {
    if(delta < 0) {
    	return Math.PI * 2 + delta
    } else {
    	return -Math.PI * 2 + delta
    }
  }
  return delta;
}

function smoothCubic(x) { //кубическая функция - в отрезке 0-1 возвращает сглаженную функцию 0-1
	return x*x*(-2*x + 3);
}

//// ИСПРАВИТЬ %
function pseudoSin(x) {
	x = modPositive(x, 1.0);
	if (x > 0.5) return 16 * (x-1) * (x-0.5);
	return 16 * (-x)*(x-0.5);
}

function modPositive(x, y) {
	if ((x % y) < 0) return x % y + y;
	return x % y;
}

function isVecEqualsFloat(vec1, x, y, z) {
	return vec1.x() == x && vec1.y() == y && vec1.z() == z;
}

function isVecEquals(vec1, vec2) {
	return vec1.x() == vec2.x() && vec1.y() == vec2.y() && vec1.z() == vec2.z();
}

function cos(vec1, vec2) {
	return dot(vec1, vec2) / (len(vec1) * len(vec2));

}

function dot(vec1, vec2) {
	return vec1.x() * vec2.x() + vec1.y() * vec2.y() + vec1.z() * vec2.z();
}

function len(vec1) {
	return Math.sqrt(vec1.x() * vec1.x() + vec1.y() * vec1.y() + vec1.z() * vec1.z()); 
}

function lenSq(vec1) {
	return vec1.x() * vec1.x() + vec1.y() * vec1.y() + vec1.z() * vec1.z(); 
}

function wipFunc(x) {
	return 0.5 * pseudoSin(x + 0.75) + 0.5;
}